from datetime import datetime
from decimal import Decimal
from io import BytesIO

import numpy as np  # type: ignore
import pyarrow  # type: ignore
import pyarrow as pa  # type: ignore


class ArbitraryBinaryArray:
    def __init__(self, val):
        self.val = val

    def __arrow_array__(self, type=None):
        # convert the underlying array values to a pyarrow Array
        mem_file = BytesIO()
        np.save(mem_file, self.val)  # noqa
        return pa.array([bytearray(mem_file.getvalue())], pa.binary())


DESERIALIZE_SEQUENCE_OPS = {
    "SELECT_column": lambda x: getattr(x, "column")(DEFAULT_SERIALIZED_COLUMN_NAME),
    "NATIVE_dict": lambda x: x.to_pandas().to_dict("records"),
    "NATIVE_list": lambda x: x.to_pylist(),
    "FINAL_list": lambda x: list(x),
    "FINAL_tuple": lambda x: tuple(x),
    "FINAL_first": lambda x: x[0],
}

NATIVE_TO_ARROW = {
    int: pyarrow.int8(),
    list: pyarrow.list_,
    float: pyarrow.float64(),
    Decimal: pyarrow.decimal128,
    str: pyarrow.string(),
    dict: pyarrow.struct,
    datetime: pyarrow.timestamp,
}

DEFAULT_SERIALIZED_COLUMN_NAME = "SerializedColumn_gen"



import typing as t
from enum import Enum, IntEnum

import numpy as np
import pandas as pd
import pytest
from pydantic import BaseModel
from taktile_types.enums.monitor import MonitorSourceType, MonitorType
from taktile_types.schemas.monitor import MonitorDataV2, MonitoringPayload

from tktl.future.monitor import Tracker, extract_monitor_element


class Model3(BaseModel):
    a: int
    b: float
    c: bool
    d: str


class Model2(BaseModel):
    a: t.Optional[int]
    b: t.Optional[float]
    c: t.Optional[bool]
    d: t.Optional[str]
    e: Model3


class Model1(BaseModel):
    a: int
    b: float
    c: bool
    d: str
    e: Model2


TEST_INSTANCE = Model1(
    a=1, b=2.0, c=True, d="string", e=Model2(e=Model3(a=5, b=6.0, c=False, d="string2"))
)


@pytest.mark.parametrize(
    "column, key, result, type_",
    [
        (pd.Series([1, 2, 3], dtype="int64"), "test", [1, 2, 3], MonitorType.NUMERIC),
        (
            pd.Series([1.0, 2.0, 3.0], dtype="float64"),
            "test",
            [1.0, 2.0, 3.0],
            MonitorType.NUMERIC,
        ),
        (
            pd.Series(["c", "", None], dtype="object"),
            "test",
            ["c", "", None],
            MonitorType.CATEGORY,
        ),
        (
            pd.Series([True, False, True], dtype="bool"),
            "test",
            [True, False, True],
            MonitorType.CATEGORY,
        ),
        (
            pd.Series(["a", "b", None], dtype="category"),
            "test",
            ["a", "b", np.nan],
            MonitorType.CATEGORY,
        ),
        (
            pd.Series([1, 2, 3], dtype="category"),
            "test",
            [1, 2, 3],
            MonitorType.CATEGORY,
        ),
        (
            pd.Series(["2020-01-01", None, "2020-01-03"], dtype="datetime64[ns]"),
            "test",
            [1577836800, None, 1578009600],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "a",
            [1],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "b",
            [2.0],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "c",
            [True],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "d",
            ["string"],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "e.a",
            [None],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "e.b",
            [None],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "e.c",
            [None],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "e.d",
            [None],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "e.e.a",
            [5],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "e.e.b",
            [6.0],
            MonitorType.NUMERIC,
        ),
        (
            TEST_INSTANCE,
            "e.e.c",
            [False],
            MonitorType.CATEGORY,
        ),
        (
            TEST_INSTANCE,
            "e.e.d",
            ["string2"],
            MonitorType.CATEGORY,
        ),
    ],
)
def test_monitor_extraction(column, key, result, type_):
    assert extract_monitor_element(column, key) == (result, type_)


def test_wrong_key(caplog):
    res = extract_monitor_element(TEST_INSTANCE, "e.e.not_a_key")
    assert res is None
    assert (
        "KeyError while extracting pydantic value: "
        "'Model3' object has no attribute 'not_a_key'"
    ) in caplog.text


class MyStrEnum(str, Enum):
    one = "one"
    two = "two"


class MyIntEnum(IntEnum):
    one = 1
    two = 2


class ModelE3(BaseModel):
    a: MyStrEnum

    class Config:
        # with this attribute set, MyIntEnum().a is of type str
        # without, it is of type MyStrEnum
        use_enum_values = True


class ModelE2(BaseModel):
    a: MyIntEnum


class ModelE1(BaseModel):
    a: MyStrEnum
    b: ModelE2
    c: ModelE3
    d: int


ENUM_TEST_INSTANCE = ModelE1(
    a="one",
    b=ModelE2(a=2),
    c=ModelE3(a="one"),
    d=42,
)


@pytest.mark.parametrize(
    "column, key, expected_result, expected_type",
    [
        (
            ENUM_TEST_INSTANCE,
            "a",
            ["one"],
            MonitorType.CATEGORY,
        ),
        (
            ENUM_TEST_INSTANCE,
            "b.a",
            [2],
            MonitorType.CATEGORY,
        ),
        (
            ENUM_TEST_INSTANCE,
            "c.a",
            ["one"],
            MonitorType.CATEGORY,
        ),
        (
            ENUM_TEST_INSTANCE,
            "d",
            [42],
            MonitorType.NUMERIC,
        ),
    ],
)
def test_monitor_extraction_enum(column, key, expected_result, expected_type):
    result, type_ = extract_monitor_element(column, key)
    assert (expected_result, expected_type) == (result, type_)
    assert not isinstance(result[0], Enum)


def test_tracker():
    class FakeRequest:
        headers = {"user-agent": "test-agent"}

    tracker = Tracker(FakeRequest(), endpoint_name="test")

    tracker.log_categorical("key1", "value1")
    tracker.log_numerical("key2", 2)

    final_payload = tracker.finalize_payload()

    assert final_payload == MonitoringPayload(
        version=2,
        data=[
            MonitorDataV2(
                value=["value1"],
                type=MonitorType.CATEGORY,
                name="key1",
                source_type=MonitorSourceType.CUSTOM,
            ),
            MonitorDataV2(
                value=[2],
                type=MonitorType.NUMERIC,
                name="key2",
                source_type=MonitorSourceType.CUSTOM,
            ),
        ],
        timestamp=final_payload.timestamp,
        user_agent="test-agent",
        endpoint="test",
        git_sha="unknown",
        git_ref="unknown",
        repository_id="unknown",
    )


def test_tracker_null():
    class FakeRequest:
        headers = {"user-agent": "test-agent"}

    tracker = Tracker(FakeRequest(), endpoint_name="test")

    tracker.log_categorical("key1", ["value1", float("nan")])
    tracker.log_numerical("key2", [2, float("nan")])

    final_payload = tracker.finalize_payload()

    assert final_payload == MonitoringPayload(
        version=2,
        data=[
            MonitorDataV2(
                value=["value1", None],
                type=MonitorType.CATEGORY,
                name="key1",
                source_type=MonitorSourceType.CUSTOM,
            ),
            MonitorDataV2(
                value=[2, None],
                type=MonitorType.NUMERIC,
                name="key2",
                source_type=MonitorSourceType.CUSTOM,
            ),
        ],
        timestamp=final_payload.timestamp,
        user_agent="test-agent",
        endpoint="test",
        git_sha="unknown",
        git_ref="unknown",
        repository_id="unknown",
    )



from typing import Optional, Union

from pyarrow.flight import FlightClient  # type: ignore
from taktile_client.http_client import API
from taktile_types.enums.instance import ServiceType

from tktl.core.clients.taktile import DeploymentApiClient
from tktl.core.loggers import LOG, Logger, set_verbosity
from tktl.core.schemas.repository import Endpoint
from tktl.login import login


class Client(object):
    TRANSPORT: ServiceType

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str,
        endpoint_name: str,
        local: bool = False,
        logger: Logger = LOG,
        verbosity: int = 0,
    ):
        set_verbosity(verbosity)
        login(api_key)
        self.local = local
        self.taktile_client = DeploymentApiClient(logger=logger)
        self.repository_name = repository_name
        self.branch_name = branch_name
        self.endpoint_name = endpoint_name
        self.logger = logger
        self._location: Optional[str] = None
        self._endpoint = None
        self._client: Union[None, API, FlightClient] = None
        self._endpoint_has_rest_sample_data: Optional[bool] = False
        self._endpoint_has_arrow_sample_data: Optional[bool] = False

    @property
    def local_endpoint(self) -> str:
        raise NotImplementedError

    @classmethod
    def for_url(cls, api_key: str, url: str, endpoint_name: str):
        raise NotImplementedError

    @staticmethod
    def format_url(url: Optional[str]) -> str:
        raise NotImplementedError

    @property
    def endpoint_has_rest_sample_data(self) -> Optional[bool]:
        return self._endpoint_has_rest_sample_data

    @endpoint_has_rest_sample_data.setter
    def endpoint_has_rest_sample_data(self, value: bool) -> None:
        self._endpoint_has_rest_sample_data = value

    @property
    def endpoint_has_arrow_sample_data(self) -> Optional[bool]:
        return self._endpoint_has_arrow_sample_data

    @endpoint_has_arrow_sample_data.setter
    def endpoint_has_arrow_sample_data(self, value: bool):
        self._endpoint_has_arrow_sample_data = value

    @property
    def client(self) -> Union[API, FlightClient]:
        return self._client

    @client.setter
    def client(self, client: Union[API, FlightClient]):
        self._client = client

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        self._location = value

    def set_client_and_location(self, location: str, client: Union[API, FlightClient]):
        self._location = location
        self._client = client

    def get_deployment_location(self):
        deployment = self.taktile_client.get_deployment_by_branch_name(
            repository_name=self.repository_name,
            branch_name=self.branch_name,
        )
        return self.format_url(deployment.public_docs_url)

    def get_endpoint_and_location(self) -> str:
        if self.local:
            location = self.local_endpoint
        else:
            location = self._get_endpoint_location()
        return location

    def _set_data_available(self, endpoint: Endpoint) -> None:
        self.endpoint_has_rest_sample_data = endpoint.has_rest_sample_data
        self.endpoint_has_arrow_sample_data = endpoint.has_arrow_sample_data

    def _get_endpoint_location(self) -> str:
        deployment, endpoint = self.taktile_client.get_endpoint_by_name(
            repository_name=self.repository_name,
            endpoint_name=self.endpoint_name,
            branch_name=self.branch_name,
        )
        self._set_data_available(endpoint=endpoint)
        return self.format_url(deployment.public_docs_url)

    def list_deployments(self):
        raise NotImplementedError

    def get_sample_data(self):
        raise NotImplementedError

    def get_schema(self, *args, **kwargs):
        raise NotImplementedError

    def health(self):
        raise NotImplementedError



import logging
import pathlib
from typing import Optional, Union

import yaml

from tktl import __version__
from tktl.core.config import Settings, settings
from tktl.core.managers.constants import CONFIG_FILE_TEMPLATE

log = logging.getLogger("root")


class BaseConfigManager(object):
    """Base class for managing a configuration file.
    Based on the amazing github.com/polyaxon/polyaxon config manager setup
    """

    TKTL_DIR: Optional[str]
    CONFIG_FILE_NAME: str
    CONFIG: Optional[dict]
    SETTINGS: Settings = settings

    @staticmethod
    def create_dir(dir_path: str) -> None:
        try:
            pathlib.Path(dir_path).mkdir(parents=True, exist_ok=True)
        except OSError as e:
            # Except permission denied and potential race conditions
            # in multi-threaded environments.
            log.error(f"Could not create config directory {dir_path}: {repr(e)}")

    @classmethod
    def get_config_file_path(cls) -> pathlib.Path:
        if not cls.TKTL_DIR:
            # local to this directory
            base_path = pathlib.Path.cwd()
        else:
            base_path = pathlib.Path(cls.TKTL_DIR)
        return base_path / cls.CONFIG_FILE_NAME

    @classmethod
    def init_config(cls, init: Union[str, bool] = False) -> None:
        cls.set_config(init=init)

    @classmethod
    def is_initialized(cls) -> bool:
        config_file_path = cls.get_config_file_path()
        return config_file_path.is_file()

    @classmethod
    def set_config(cls, init: Union[str, bool] = False) -> None:
        config_file_path = cls.get_config_file_path()

        if config_file_path.is_file() and init:
            log.debug(
                "%s file already present at %s", cls.CONFIG_FILE_NAME, config_file_path
            )
            return

        if init:
            config_file_content = CONFIG_FILE_TEMPLATE.format(version=__version__)
            with open(config_file_path, "w") as config_file_handle:
                config_file_handle.write(config_file_content)

    @classmethod
    def get_config(cls) -> Optional[dict]:
        if not cls.is_initialized():
            return None
        config_file_path = cls.get_config_file_path()
        with open(config_file_path, "r") as config_file:
            return yaml.safe_load(config_file)

    @classmethod
    def get_value(cls, key: str) -> Union[Optional[str], Optional[dict]]:
        config = cls.get_config()
        if config:
            if key in config.keys():
                return config[key]
            else:
                log.warning("Config `%s` has no key `%s`", cls.CONFIG_FILE_NAME, key)
        return None



import typing as t
from functools import singledispatch

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .exceptions import ValidationError


@singledispatch
def validate(
    value: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    *,
    sample: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    minimal_subset: t.Optional[t.Set[str]] = None,
) -> t.Union[np.ndarray, pd.DataFrame, pd.Series]:
    raise ValidationError(f"Can't validate value of type: {type(value)}")


@validate.register
def _validate_numpy(
    value: np.ndarray,
    *,
    sample: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    minimal_subset: t.Optional[t.Set[str]] = None,
) -> np.ndarray:
    if type(value) != type(sample):
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    if value.shape[1:] != sample.shape[1:]:
        raise ValidationError(
            "Could not validate numpy array shape. "
            f"Value has shape {value.shape[1:]}, expected shape: {sample.shape[1:]}"
        )

    # We choose not to validate the dtypes here

    return value


@validate.register
def _validate_series(
    value: pd.Series,
    *,
    sample: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    minimal_subset: t.Optional[t.Set[str]] = None,
) -> pd.Series:
    if type(value) != type(sample):
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    # We choose not to validate the dtypes here

    return value


@validate.register
def _validate_dataframe(
    value: pd.DataFrame,
    *,
    sample: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    minimal_subset: t.Optional[t.Set[str]] = None,
) -> pd.DataFrame:
    if not isinstance(sample, pd.DataFrame):
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    # We choose not to validate the dtypes here

    def decode(x: t.Union[str, bytes]) -> str:
        if isinstance(x, bytes):
            return x.decode()
        return x

    value.columns = [decode(x) for x in value.columns]

    sent_columns = {x for x in value.columns.to_list()}  # type: ignore
    expected_columns = {x for x in sample.columns.to_list()}  # type: ignore

    if minimal_subset is None:
        if sent_columns != expected_columns:
            missing = expected_columns.difference(sent_columns)
            superfluous = sent_columns.difference(expected_columns)
            raise ValidationError(
                "Column mismatch: " f"Missing columns: {missing} "
                if missing
                else "" f"Superfluous columns: {superfluous}"
                if superfluous
                else ""
            )
    else:
        if not sent_columns.issubset(expected_columns):
            raise ValidationError(
                "Column mismatch: "
                f"Superfluous columns: {sent_columns.difference(expected_columns)}."
            )
        if not minimal_subset.issubset(sent_columns):
            raise ValidationError(
                "Column mismatch: "
                f"Missing columns: {minimal_subset.difference(sent_columns)}."
            )
        sample = sample[value.columns]

    value = value[sample.columns.to_list()]  # type: ignore # ordering
    return value.astype(sample.dtypes.to_dict())  # type: ignore # dtypes. TODO: Improve this



import pathlib
import pkgutil
import sys
from typing import Optional

from tktl import __version__
from tktl.core.loggers import LOG
from tktl.core.managers.base import BaseConfigManager
from tktl.core.managers.constants import TEMPLATE_PROJECT_DIRS, TEMPLATE_PROJECT_FILES
from tktl.core.t import TemplateT
from tktl.core.validation.project import validate_project_contents


class ProjectManager(BaseConfigManager):
    """Manages project configuration tktl.yaml file."""

    TKTL_DIR: Optional[str] = None
    CONFIG_FILE_NAME: str = "tktl.yaml"
    CONFIG: Optional[dict] = None

    @classmethod
    def set_path(cls, path: str) -> None:
        cls.TKTL_DIR = path

    @classmethod
    def init(cls, path: str, name: str, template: TemplateT) -> None:
        cls.create_dir(path)
        cls.set_path(path)
        cls.init_config(name)
        cls.create_template(template)

    @classmethod
    def init_project(cls, path: Optional[str], name: str, template: TemplateT) -> str:
        if not path:
            _path = pathlib.Path.cwd()
        else:
            _path = pathlib.Path(path)
        project_path = _path / name
        if project_path.is_dir():
            cls.safe_init(str(project_path), name, template)
        else:
            cls.init(str(project_path), name, template)
        return str(project_path)

    @classmethod
    def safe_init(cls, path: str, name: str, template: TemplateT) -> None:
        LOG.log(
            "WARNING: Path specified already exists. "
            "Any configuration will be overwritten. Proceed?[Y/n]",
            color="yellow",
        )
        overwrite = input()
        if overwrite == "Y":
            cls.init(path, name, template)
        elif overwrite == "n":
            sys.exit()
        else:
            cls.safe_init(path, name, template)

    @classmethod
    def create_template(cls, template: TemplateT) -> None:
        cls.create_subdirs()
        cls.copy_dist_files(template)

    @classmethod
    def copy_dist_files(cls, template: TemplateT) -> None:
        for file in TEMPLATE_PROJECT_FILES[template]:
            file = str(pathlib.Path("templates", template.value, file))
            data = pkgutil.get_data("tktl", file)
            if data:
                cls._copy_file(file, data, template)

    @classmethod
    def create_subdirs(cls) -> None:
        for path in TEMPLATE_PROJECT_DIRS:
            full_path = (
                pathlib.Path(cls.TKTL_DIR) / path
                if cls.TKTL_DIR
                else pathlib.Path(path)
            )
            full_path.mkdir(parents=True, exist_ok=True)

    @classmethod
    def _copy_file(cls, file, data, template, replace_path=""):
        strip = f"templates/{template.value}/"
        file_path = file.replace(strip, replace_path)
        tktl_dir = pathlib.Path(cls.TKTL_DIR)
        py_major, py_minor = sys.version_info.major, sys.version_info.minor
        docker_tag_suffix = f"{__version__}-py{py_major}.{py_minor}"
        try:
            with open(tktl_dir / file_path, "w") as f:
                f.write(data.decode("utf-8").replace("__version__", docker_tag_suffix))
        except UnicodeDecodeError:
            with open(tktl_dir / file_path, "wb") as f:
                f.write(data)

    @classmethod
    def validate_project_config(cls, path="."):
        return validate_project_contents(path=path)



import pathlib
import typing as t

import pandas as pd  # type: ignore
from taktile_types.enums.endpoint import EndpointKinds, ProfileKinds

from .abc import XType, YType
from .arrow import ArrowEndpoint


class ProfiledEndpoint(ArrowEndpoint):
    kind: EndpointKinds = EndpointKinds.PROFILED

    @property
    def input_names(self):
        return self.profile_columns

    @property
    def output_names(self):
        return self.y.name

    @property
    def profile_columns(self) -> t.Optional[t.List[str]]:
        return (
            self._profile_columns
            if self._profile_columns is not None
            else self.X.columns.to_list()
        )

    @staticmethod
    def supported(
        *,
        X: XType = None,
        y: YType = None,
        profile: t.Optional[str] = None,
    ) -> bool:

        if profile not in ProfileKinds.set():
            return False

        strict = isinstance(X, pd.DataFrame) and isinstance(y, pd.Series)
        lazy = isinstance(X, pathlib.Path) and isinstance(y, pathlib.Path)

        return strict or lazy



import pathlib
import shutil

import pytest

from tktl.commands.init import init_project
from tktl.commands.validate import validate_project_config
from tktl.core.managers.constants import TEMPLATE_PROJECT_DIRS, TEMPLATE_PROJECT_FILES
from tktl.core.managers.project import ProjectManager
from tktl.core.t import TemplateT


def test_create_in_existing_directory(monkeypatch, tmpdir):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(
        tmpdir.dirpath(), "test_project_dir", TemplateT.REPAYMENT
    )
    for d in TEMPLATE_PROJECT_DIRS:
        assert (pathlib.Path(tmpdir.dirpath()) / "test_project_dir" / d).is_dir()
    for file in TEMPLATE_PROJECT_FILES[TemplateT.REPAYMENT]:
        assert (pathlib.Path(tmpdir.dirpath()) / "test_project_dir" / file).exists()
    assert (pathlib.Path(tmpdir.dirpath()) / "test_project_dir" / "tktl.yaml").exists()


def test_create_project():
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    sample_project_dir = pathlib.Path("sample_project")
    for file in TEMPLATE_PROJECT_FILES[TemplateT.REPAYMENT]:
        assert (sample_project_dir / file).is_file()
    assert (sample_project_dir / "tktl.yaml").exists()

    for d in TEMPLATE_PROJECT_DIRS:
        assert (sample_project_dir / d).is_dir()
    shutil.rmtree("sample_project")


def test_safe_init(monkeypatch, caplog):
    monkeypatch.setattr("builtins.input", lambda: "n")

    with pytest.raises(SystemExit):
        ProjectManager.safe_init(None, "sample_project", TemplateT.REPAYMENT)
        assert "Path specified already exists" in caplog.text


def test_overwrite(monkeypatch, capsys):
    monkeypatch.setattr("builtins.input", lambda: "Y")
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    init_project(None, "sample_project", TemplateT.REPAYMENT)
    out, err = capsys.readouterr()
    assert "Project scaffolding created successfully" in out
    shutil.rmtree("sample_project")


def test_validate(capsys):
    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding is valid!\n" in out
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    shutil.rmtree("sample_project/src")
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert err is not None
    assert "Project scaffolding is invalid: Missing Files or Directory ❌\n" in out
    assert "Missing directories in repository: src" in err
    shutil.rmtree("sample_project")

    ProjectManager.init_project(None, "sample_project", TemplateT.REPAYMENT)
    with open("sample_project/tktl.yaml", "w") as inf:
        inf.write(
            """
default_branch_name: master
service:
  replicas: 3
  compute_type: CPUqqqqqqqq
  endpoint_type: RESTuuuuuu
"""
        )
    validate_project_config("sample_project")
    out, err = capsys.readouterr()
    assert "Project scaffolding is invalid: Invalid Config File ❌\n" in out
    shutil.rmtree("sample_project")



import logging

import click

from tktl.commands.init import init_project
from tktl.core.t import TemplateT

_logger = logging.getLogger("root")


@click.command(help="Creates project scaffolding")
@click.option(
    "--name", help="Name of the project", default="tktl-serving", required=True
)
@click.option(
    "--path", help="Directory where the new project will be created", required=False
)
@click.option(
    "--template",
    help="Initial template of the new project",
    type=click.Choice(TemplateT.list(), case_sensitive=False),
    default="repayment",
)
def init(path: str, name: str, template: TemplateT) -> None:
    """Creates a new project with the necessary scaffolding.

    Creates the supporting files needed. The directory structure of a new
    project , and the files within it will look like this:

        .dockerignore
        .gitattributes
        .gitignore
        .buildfile
        README.md
        assets              # Where your ML models and test data live
        requirements.txt    # User-specified requirements
        src                 # Source code for endpoint definitions
        tests               # User-specified tests
        tktl.yaml           # Taktile configuration options

    """
    init_project(path=path, name=name, template=TemplateT[template.upper()])



from tktl.core import ExtendedEnum


class TemplateT(str, ExtendedEnum):
    REPAYMENT = "repayment"
    BARE = "bare"
    PRECOMPUTED = "precomputed"
    MONITORING = "monitoring"


class UserRepoConfigFileNameT(str, ExtendedEnum):
    YAML = "tktl.yaml"
    YML = "tktl.yml"


class UserProjectFileT(str, ExtendedEnum):
    FILE = "file"
    DIRECTORY = "dir"


class RequiredUserProjectPathsT(str, ExtendedEnum):
    SRC = "src"
    ASSETS = "assets"
    TESTS = "tests"
    TKTL_YML = "tktl.yml"
    TKTL_YAML = "tktl.yaml"
    DOCKERFILE = ".buildfile"
    REQS = "requirements.txt"

    @classmethod
    def strictly_required_files(cls):
        return {cls.DOCKERFILE.value, cls.REQS.value}

    @classmethod
    def strictly_required_dirs(cls):
        return {cls.SRC.value, cls.ASSETS.value}


class DeploymentStateFailed(str, ExtendedEnum):
    BUILDING_FAILED = "building_failed"
    DEPLOYING_FAILED = "deploying_failed"
    ROUTING_FAILED = "routing_failed"
    PROFILING_FAILED = "profiling_failed"
    ERRORED = "errored"


class DeploymentStateSucceed(str, ExtendedEnum):
    RUNNING = "running"


class DeploymentStatePending(str, ExtendedEnum):
    BUILDING = "building"
    DEPLOYING = "deploying"
    ROUTING = "routing"
    PROFILING = "profiling"


class DeploymentStatesT(str, ExtendedEnum):

    BUILDING = "building"
    DEPLOYING = "deploying"
    ROUTING = "routing"
    PROFILING = "profiling"
    RUNNING = "running"
    DELETED = "deleted"

    BUILDING_FAILED = "building_failed"
    DEPLOYING_FAILED = "deploying_failed"

    ROUTING_FAILED = "routing_failed"
    PROFILING_FAILED = "profiling_failed"

    # Error after entering the running state
    ERRORED = "errored"


class ProjectAssetT(str, ExtendedEnum):
    DATA = "data"
    MODEL = "model"


class ProjectAssetSourceT(str, ExtendedEnum):
    S3 = "s3"
    LOCAL = "local"
    LFS = "lfs"


class EndpointKinds(str, ExtendedEnum):
    CUSTOM = "custom"
    TABULAR = "tabular"
    BINARY = "binary"
    REGRESSION = "regression"
    MULTICLASS = "multiclass"


class RestSchemaTypes(ExtendedEnum):
    DICT = "Dict"
    SEQUENCE = "Sequence"
    ARRAY = "Array"
    FLAT_ARRAY = "FlatArray"
    SERIES = "Series"
    DATAFRAME = "DataFrame"
    SINGLE_VALUE = "SingleValue"
    CUSTOM_MODEL = "CustomModel"
    ANY = "Any"


class LazifyType(str, ExtendedEnum):
    """Type of lazification to be performed"""

    SHAP = "Shap"
    FIRST_VALID = "FirstValid"



import pathlib
import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from taktile_client.rest.serialize import deserialize, serialize, to_example
from taktile_types.enums.endpoint import EndpointKinds

from tktl.future.registration.exceptions import ValidationError
from tktl.future.registration.validation import validate
from tktl.future.utils import JSONStructure

from .abc import XType, YType
from .typed import TypedEndpoint


class ArrowEndpoint(TypedEndpoint):
    kind: EndpointKinds = EndpointKinds.ARROW

    @staticmethod
    def supported(
        *,
        X: XType = None,
        y: YType = None,
        profile: t.Optional[str] = None,
    ) -> bool:

        if profile is not None:
            return False

        strict = isinstance(X, (np.ndarray, pd.DataFrame, pd.Series)) and isinstance(
            y, (np.ndarray, pd.DataFrame, pd.Series)
        )

        lazy = isinstance(X, pathlib.Path) and isinstance(y, pathlib.Path)

        return strict or lazy

    def deserialize_function(
        self,
    ) -> t.Callable[[str], t.Union[pd.Series, pd.DataFrame, np.ndarray]]:
        def _deserialize(serial_value, explainer: bool = False):

            minimal_subset = None

            if explainer:
                minimal_subset = (
                    set(self._profile_columns) if self._profile_columns else None
                )

            try:
                value = deserialize(self._X, value=serial_value)
                value = validate(
                    value,
                    sample=self._X,
                    minimal_subset=minimal_subset,
                )
                return value
            except ValidationError as exc:
                raise ValidationError(f"Validation error on input: {str(exc)}") from exc

        return _deserialize

    def serialize_function(self) -> t.Callable[[t.Any], str]:
        def _serialize(value):
            try:
                value = validate(value, sample=self._y)
                return serialize(value)
            except ValidationError as exc:
                raise ValidationError(
                    f"Validation error on output: {str(exc)}"
                ) from exc

        return _serialize

    def request_type(self) -> object:
        return JSONStructure

    def request_example(self) -> str:
        return to_example(self._X)

    def response_type(self) -> object:
        return JSONStructure

    def response_example(self) -> str:
        return to_example(self._y)



import os

import pytest
import requests
from pydantic import BaseModel

from tktl.core.clients import API
from tktl.core.config import set_api_key, settings


@pytest.mark.skipif("TEST_USER_API_KEY" not in os.environ, reason="No User Logged In")
def test_instantiate_api_client():
    key = os.environ["TEST_USER_API_KEY"]
    set_api_key(key)
    client = API(api_base=settings.DEPLOYMENT_API_URL, api_key=key)
    assert "X-Api-Key" in client._headers
    assert client._headers["X-Api-Key"] == key
    assert client._base == f"{settings.DEPLOYMENT_API_URL}"


@pytest.mark.parametrize("verb", [("post"), ("get"), ("patch"), ("put"), ("delete")])
def test_request_failures(verb: str):
    client = API(api_base=settings.DEPLOYMENT_API_URL)
    with pytest.raises(requests.exceptions.RequestException):
        client.call(verb=verb, path="non/existent/path")


def test_interpret_response():
    client = API(api_base=settings.DEPLOYMENT_API_URL)

    with pytest.raises(requests.exceptions.HTTPError):
        client.call(verb="post", path="non/existent/path", model=BaseModel)

    class TestModel(BaseModel):
        id: int
        title: str

    client = API(api_base="https://my-json-server.typicode.com/typicode/demo/posts/1")
    response = client.call(verb="get", path="", model=TestModel)

    assert response.dict()["id"] == 1



import pathlib
import traceback
import typing as t

from click import ClickException
from docker.errors import APIError  # type: ignore
from pydantic import ValidationError

from tktl.core.exceptions import (
    MissingDocker,
    NoContentsFoundException,
    UserRepoValidationException,
)
from tktl.core.loggers import LOG
from tktl.core.managers.docker import DockerManager
from tktl.core.managers.project import ProjectManager
from tktl.core.schemas.project import ProjectValidationOutput
from tktl.core.validation.outputs import (
    ConfigFileValidationFailure,
    ProjectValidationFailure,
)


def get_docker_manager(path: str) -> DockerManager:
    try:
        return DockerManager(path)
    except MissingDocker:
        raise ClickException(
            "Couldn't locate docker. Please make sure it is installed, "
            "or use the DOCKER_HOST environment variable."
        )


def validate_project_config(path: str):
    try:
        LOG.log("=== VALIDATE PROJECT CONFIG STEP ===")
        ProjectManager.validate_project_config(path)
    except ValidationError as e:
        validation_output = ProjectValidationOutput(
            title=ConfigFileValidationFailure.title,
            summary=ConfigFileValidationFailure.summary,
            text=ConfigFileValidationFailure.format_step_results(validation_errors=e),
        )
        _log_failure(validation_output)
        return
    except (NoContentsFoundException, UserRepoValidationException) as e:
        validation_output = ProjectValidationOutput(
            title=ProjectValidationFailure.title,
            summary=ProjectValidationFailure.summary,
            text=ProjectValidationFailure.format_step_results(validation_errors=e),
        )
        _log_failure(validation_output)
        return
    LOG.log("Project scaffolding is valid!", color="green")


def _log_failure(validation_output: ProjectValidationOutput):
    LOG.log(f"Project scaffolding is invalid: {validation_output.title}", color="red")
    LOG.log(validation_output.summary, color="red", err=True)
    LOG.log(validation_output.text, color="red", err=True)


def build_image(
    dm: DockerManager,
    path: str,
    cache: bool,
    prune: bool,
    secrets: t.Optional[t.Dict[str, str]] = None,
) -> str:
    LOG.log("=== BUILD STEP ===")
    abs_path = (pathlib.Path(path) / ".buildfile").resolve()
    if secrets is None:
        secrets = {}
    buildargs = {
        "tktl_local": "true",
        **secrets,
    }

    image = dm.build_image(
        dockerfile=str(abs_path), cache=cache, prune=prune, buildargs=buildargs
    )

    return image


def status_output(status, allowed_status_codes: t.List[int] = [0]):
    if status["StatusCode"] in allowed_status_codes:
        LOG.log("Success", color="green")
    else:
        LOG.log("Error", color="red")
        exit(1)


def validate_import(
    dm: DockerManager,
    image: str,
    secrets: t.Optional[t.Dict[str, str]] = None,
):
    LOG.log("=== VALIDATE IMPORT STEP ===")
    status, _ = dm.test_import(image, secrets=secrets)
    status_output(status)


def validate_endpoint_sample(
    dm: DockerManager,
    image: str,
    secrets: t.Optional[t.Dict[str, str]] = None,
):
    LOG.log("=== VALIDATE ENDPOINT SAMPLE DATA STEP ===")
    status, _ = dm.test_endpoint_sample(image, secrets=secrets)
    status_output(status)


def validate_unittest(
    dm: DockerManager,
    image: str,
    secrets: t.Optional[t.Dict[str, str]] = None,
):
    LOG.log("=== VALIDATE UNITTEST STEP ===")
    status, _ = dm.test_unittest(image, secrets=secrets)
    status_output(status, [0, 5])


def validate_integration(
    dm: DockerManager,
    image: str,
    timeout: int,
    retries: int,
    secrets: t.Optional[t.Dict[str, str]] = None,
):
    LOG.log("=== VALIDATE INTEGRATION STEP ===")
    LOG.log("Waiting for service to start...")
    try:
        dm.run_and_check_health(
            image,
            kill_on_success=True,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
            secrets=secrets,
        )
        LOG.log("Success", color="green")
    except:  # noqa
        LOG.log("Unable to run container. See stack trace for more info", color="red")
        traceback.print_exc()
        exit(1)


def validate_profiling(
    dm: DockerManager,
    image: str,
    timeout: int,
    retries: int,
    prune: bool,
    secrets: t.Optional[t.Dict[str, str]] = None,
):
    LOG.log("=== VALIDATE PROFILING STEP ===")
    LOG.log("Initiating service...")
    try:
        arrow_container, rest_container, = dm.run_and_check_health(
            image,
            kill_on_success=False,
            auth_enabled=False,
            timeout=timeout,
            retries=retries,
            secrets=secrets,
        )
    except Exception:
        raise ClickException(
            "Failed to run service container. "
            "Ensure service can run with `tktl validate integration`"
        )
    try:
        LOG.log("Initiating remote profiling...")
        status, container = dm.run_profiling_container()
        status_output(status)
        if prune:
            dm.remove_image(image=container.image.id)
    finally:
        try:
            arrow_container.kill()
            rest_container.kill()
            if prune:
                dm.remove_image(image=rest_container.image.id)
                dm.remove_image(image=arrow_container.image.id)
        except APIError:
            pass


def _validate_rest_response(rest_response):
    if rest_response is None:
        LOG.log("Could not access REST endpoint", color="red")
    elif rest_response.status_code != 204:
        LOG.log(f"Response status code {rest_response.status_code}", color="red")
    else:
        return True
    return False


def _validate_grpc_response(grpc_response):
    if grpc_response is None:
        LOG.log("Could not access gRPC endpoint", color="red")
    else:
        return True
    return False


def _validate_container_response(rest_response, grpc_response):
    return _validate_rest_response(
        rest_response=rest_response
    ) and _validate_grpc_response(grpc_response=grpc_response)



import os

import pytest

from tktl.core.clients import DeploymentApiClient
from tktl.core.clients.future.taktile import (
    DeploymentApiClient as FutureDeploymentApiClient,
)
from tktl.core.config import set_api_key
from tktl.core.exceptions import TaktileSdkError


@pytest.mark.skipif("TEST_USER_API_KEY" not in os.environ, reason="No User Logged In")
def test_instantiate_taktile_client():
    key = os.environ["TEST_USER_API_KEY"]
    set_api_key(key)
    client = DeploymentApiClient()
    with pytest.raises(TaktileSdkError):
        client.get_endpoint_by_name("admin/df.asfm", "", "")
    assert "X-Api-Key" in client._headers
    assert client._headers["X-Api-Key"] == key


@pytest.mark.skipif("TEST_USER_API_KEY" not in os.environ, reason="No User Logged In")
def test_instantiate_taktile_client_future():
    key = os.environ["TEST_USER_API_KEY"]
    set_api_key(key)
    client = FutureDeploymentApiClient(api_key=key)
    with pytest.raises(TaktileSdkError):
        client.get_endpoint_by_name("admin/df.asfm", "", "")
    assert "X-Api-Key" in client._headers
    assert client._headers["X-Api-Key"] == key



import numpy as np
from src.endpoints import X, repayment

prob = repayment(X)


def test_mean(json_metadata):
    mean_pred = np.mean(prob)

    json_metadata["section"] = "Plausibility"
    json_metadata["pass_message"] = f"Mean prediction is plausible ({mean_pred:.2f})"
    json_metadata["fail_message"] = f"Mean prediction is implausible ({mean_pred:.2f})"

    assert 0.85 < mean_pred < 0.90


def test_spread_10_90(json_metadata):
    p10 = np.quantile(prob, q=0.10)
    p90 = np.quantile(prob, q=0.90)
    spread = p90 - p10

    json_metadata["section"] = "Plausibility"
    json_metadata["pass_message"] = f"10%/90% Spread is plausible ({spread:.2f})"
    json_metadata["fail_message"] = f"10%/90% Spread is implausible ({spread:.2f})"

    assert 0.05 < spread < 0.15


def test_spread_25_75(json_metadata):
    p25 = np.quantile(prob, q=0.25)
    p75 = np.quantile(prob, q=0.75)
    spread = p75 - p25

    json_metadata["section"] = "Plausibility"
    json_metadata["pass_message"] = f"25%/75% Spread is plausible ({spread:.2f})"
    json_metadata["fail_message"] = f"25%/75% Spread is implausible ({spread:.2f})"

    assert 0.025 < spread < 0.05



import json
import os
import pathlib

import pytest
from click.testing import CliRunner

from tktl import main


@pytest.mark.skipif("TEST_USER_API_KEY" not in os.environ, reason="No User Logged In")
def test_login(user_key):
    """Test the CLI."""
    u, k = user_key
    runner = CliRunner()
    result = runner.invoke(main.login, [k])
    assert result.exit_code == 0, result.output
    assert result.output == f"Authentication successful for user: {u}\n"

    result = runner.invoke(main.login, ["FAKEAPIKEY"])
    assert result.exit_code == 0
    assert result.output == "Authentication failed: Key format is invalid\n"
    with open(pathlib.Path.home() / ".config" / "tktl" / "config.json", "r") as j:
        d = json.load(j)
    assert d["api-key"] != "FAKEAPIKEY"
    assert d["api-key"] == k



from enum import Enum
from typing import Iterable, List, Set


class ExtendedEnum(Enum):
    @classmethod
    def set(cls) -> Set:
        return {c.value for c in cls}

    @classmethod
    def list(cls) -> List:
        return [c.value for c in cls]

    @classmethod
    def names(cls: Iterable) -> List:
        return list(map(lambda c: c.name, cls))



from typing import List, Optional

from pydantic import UUID4

from tktl.commands import BaseDeploymentApiCommand
from tktl.core.schemas.deployment import DeploymentBase
from tktl.core.schemas.repository import Endpoint


class GetDeployments(BaseDeploymentApiCommand):
    def execute(
        self,
        repository_id: Optional[UUID4],
        repository_name: Optional[str],
        repository_owner: Optional[str],
        git_hash: Optional[str],
        branch_name: Optional[str],
        status_name: Optional[str],
        return_all: bool,
    ):
        return self.client.get_deployments(
            repository_id=repository_id,
            repository_name=repository_name,
            repository_owner=repository_owner,
            git_hash=git_hash,
            branch_name=branch_name,
            status_name=status_name,
            return_all=return_all,
        )


class GetRepositories(BaseDeploymentApiCommand):
    def execute(
        self,
        repository_name: str,
        repository_owner: str,
        return_all: bool = False,
    ):
        return self.client.get_repositories(
            repository_name=repository_name,
            repository_owner=repository_owner,
            return_all=return_all,
        )


class GetEndpoints(BaseDeploymentApiCommand):
    def execute(
        self,
        deployment_id: UUID4,
        repository_name: str,
        repository_owner: str,
        git_hash: str,
        branch_name: str,
        status_name: str,
        endpoint_name: str,
        endpoint_kind: str,
        return_all: bool = False,
    ) -> List[Endpoint]:
        return self.client.get_endpoints(
            deployment_id=deployment_id,
            repository_name=repository_name,
            repository_owner=repository_owner,
            git_hash=git_hash,
            branch_name=branch_name,
            status_name=status_name,
            endpoint_name=endpoint_name,
            endpoint_kind=endpoint_kind,
            return_all=return_all,
        )


class DeleteDeployment(BaseDeploymentApiCommand):
    def execute(self, deployment_id: UUID4) -> DeploymentBase:
        return self.client.delete_deployment(deployment_id=deployment_id)


class ListDeploymentsCommand(BaseDeploymentApiCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        # TODO: implement list deployments and show as table

        # with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
        #     instances = self._get_instances(**kwargs)

        raise NotImplementedError


class GetDeploymentDetails(BaseDeploymentApiCommand):
    def _get_table_data(self, instance):
        pass

    def execute(self, id_):
        # TODO: get deployment details
        raise NotImplementedError


class GetDeploymentMetricsCommand(BaseDeploymentApiCommand):
    def execute(
        self, deployment_id, start, end, interval, built_in_metrics, *args, **kwargs
    ):
        # TODO: stream metrics
        raise NotImplementedError


class StreamDeploymentMetricsCommand(BaseDeploymentApiCommand):
    def execute(self, **kwargs):
        # TODO: stream metrics
        raise NotImplementedError


class DeploymentLogsCommand(BaseDeploymentApiCommand):
    def execute(self, **kwargs):
        # TODO: get logs
        raise NotImplementedError

    def _get_log_row_string(self, id, log):
        raise NotImplementedError

    def _make_table(self, logs, id):
        raise NotImplementedError
